/*
 * Copyright 2004 - 2007 University of Cardiff.
 * Copyright 2008 aQute SARL, Adapted for bnd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package aQute.lib.signing;

import java.io.*;
import java.lang.reflect.*;
import java.security.*;
import java.security.cert.*;
import java.security.spec.*;
import java.util.*;
import java.util.jar.*;
import java.util.zip.*;

import sun.security.util.*;
import aQute.lib.base64.*;
import aQute.lib.osgi.*;

public class JarSigner {

    // the alias for the signing key, the private key to sign with,
    // and the certificate chain
    private String            alias;
    private PrivateKey        privateKey;
    private X509Certificate[] certChain;

    public JarSigner(String alias, PrivateKey privateKey,
            X509Certificate[] certChain) {
        this.alias = alias;
        this.privateKey = privateKey;
        this.certChain = certChain;
    }

    private static String updateDigest(MessageDigest digest,
            InputStream inputStream) throws IOException {
        byte[] buffer = new byte[2048];
        int read = 0;
        while ((read = inputStream.read(buffer)) > 0)
            digest.update(buffer, 0, read);
        inputStream.close();

        Base64 base64 = new Base64(digest.digest());
        return base64.toString();
    }

    // update the attributes in the manifest to have the
    // appropriate message digests. we store the new entries into

    // the entries Map and return it (we do not compute the digests
    // for those entries in the META-INF directory)
    // New
    private static Map<String, Attributes> updateManifestDigest(
            Manifest manifest, Jar jarFile, MessageDigest messageDigests[],
            Map<String, Attributes> entries) throws IOException {
        for (Map.Entry<String, Resource> jarEntry : jarFile.getResources()
                .entrySet()) {
            if (jarEntry.getKey().startsWith("META-INF"))
                continue;

            Attributes attributes = entries.get(jarEntry.getKey());
            if (attributes == null) {
                attributes = new Attributes();
                entries.put(jarEntry.getKey(), attributes);
            }
            for (MessageDigest messageDigest : messageDigests) {
                String algorithm = messageDigest.getAlgorithm();
                attributes.putValue(algorithm + "-Digest", updateDigest(
                        messageDigest, jarEntry.getValue().openInputStream()));
            }
        }
        return entries;
    }

    // a small helper function that will convert a manifest into an
    // array of bytes
    private byte[] serialiseManifest(Manifest manifest) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        manifest.write(baos);
        baos.flush();
        baos.close();
        return baos.toByteArray();

    }

    // create a signature file object out of the manifest and the
    // message digest
    private SignatureFile createSignatureFile(Manifest manifest,
            MessageDigest messageDigests[]) throws Exception {
        // construct the signature file and the signature block for
        // this manifest
        ManifestDigester manifestDigester = new ManifestDigester(
                serialiseManifest(manifest));
        return new SignatureFile(messageDigests, manifest, manifestDigester,
                this.alias, true);

    }

    // the actual JAR signing method -- this is the method which
    // will be called by those wrapping the JARSigner class
    public void signJar(Jar jar) throws Exception {

        Manifest manifest = jar.getManifest();
        Map<String, Attributes> entries = manifest.getEntries();

        // create the message digest and start updating the
        // the attributes in the manifest to contain the SHA1
        // digests
        MessageDigest sha1 = MessageDigest.getInstance("SHA1");
        MessageDigest md5 = MessageDigest.getInstance("MD5");
        MessageDigest[] messageDigests = { sha1, md5 };
        updateManifestDigest(manifest, jar, messageDigests, entries);

        // construct the signature file object and the
        // signature block objects
        SignatureFile signatureFile = createSignatureFile(manifest,
                messageDigests);
        SignatureFile.Block block = signatureFile.generateBlock(privateKey,
                certChain, true, null);

        String signatureFileName = signatureFile.getMetaName();

        ByteArrayOutputStream bos = new ByteArrayOutputStream();
        signatureFile.write(bos);
        jar.putResource(signatureFileName, new EmbeddedResource(bos
                .toByteArray(), 0));

        // write out the signature block file -- again, the block
        // will name itself appropriately
        String signatureBlockName = block.getMetaName();
        bos = new ByteArrayOutputStream();
        block.write(bos);
        jar.putResource(signatureBlockName, new EmbeddedResource(bos
                .toByteArray(), 0));
    }

    private static <T> Constructor<T> findConstructor(Class<T> c,
            Class<?>... argTypes) throws NoSuchMethodException {
        Constructor<T> ct = c.getDeclaredConstructor(argTypes);
        ct.setAccessible(true);
        return ct;
    }

    private static <T> Method findMethod(Class<T> c, String methodName,
            Class<?>... argTypes) throws NoSuchMethodException {
        Method m = c.getDeclaredMethod(methodName, argTypes);
        if (m == null) {
            throw new RuntimeException(c.getName());
        }
        m.setAccessible(true);
        return m;
    }

    private class SignatureFile {
        private Object              sigFile;
        private Class<?>            JDKsfClass;
        private Method              getMetaNameMethod;
        private Method              writeMethod;
        private static final String JDK_SIGNATURE_FILE = "sun.security.tools.SignatureFile";
        private static final String GETMETANAME_METHOD = "getMetaName";
        private static final String WRITE_METHOD       = "write";

        public SignatureFile(MessageDigest digests[], Manifest mf,
                ManifestDigester md, String baseName, boolean signManifest)
                throws ClassNotFoundException, NoSuchMethodException,
                InstantiationException, IllegalAccessException,
                InvocationTargetException {

            JDKsfClass = Class.forName(JDK_SIGNATURE_FILE);

            Constructor<?> constructor = findConstructor(JDKsfClass,
                    MessageDigest[].class, Manifest.class,
                    ManifestDigester.class, String.class, Boolean.TYPE);

            sigFile = constructor.newInstance(digests, mf, md, baseName,
                    signManifest);

            getMetaNameMethod = findMethod(JDKsfClass, GETMETANAME_METHOD);
            writeMethod = findMethod(JDKsfClass, WRITE_METHOD,
                    OutputStream.class);
        }

        public Block generateBlock(PrivateKey privateKey,
                X509Certificate[] certChain, boolean externalSF, ZipFile zipFile)
                throws Exception {
            return new Block(this, privateKey, certChain, externalSF, zipFile);
        }

        public Class<?> getJDKSignatureFileClass() {
            return JDKsfClass;
        }

        public Object getJDKSignatureFile() {
            return sigFile;
        }

        public String getMetaName() throws IllegalAccessException,
                InvocationTargetException {
            return (String) getMetaNameMethod.invoke(sigFile);
        }

        public void write(OutputStream os) throws IllegalAccessException,
                InvocationTargetException {
            writeMethod.invoke(sigFile, os);
        }

        private class Block {

            private Object              block;

            private static final String JDK_BLOCK          = JDK_SIGNATURE_FILE
                                                                   + "$Block";
            private static final String JDK_CONTENT_SIGNER = "com.sun.jarsigner.ContentSigner";

            private Method              getMetaNameMethod;
            private Method              writeMethod;

            public Block(SignatureFile sfg, PrivateKey privateKey,
                    X509Certificate[] certChain, boolean externalSF,
                    ZipFile zipFile) throws ClassNotFoundException,
                    NoSuchMethodException, InstantiationException,
                    IllegalAccessException, InvocationTargetException {

                Class<?> blockClass = Class.forName(JDK_BLOCK);

                Class<?> contentSignerClass = Class.forName(JDK_CONTENT_SIGNER);

                Constructor<?> constructor = findConstructor(blockClass, sfg
                        .getJDKSignatureFileClass(), PrivateKey.class,
                        X509Certificate[].class, Boolean.TYPE, String.class,
                        X509Certificate.class, contentSignerClass,
                        String[].class, ZipFile.class);

                getMetaNameMethod = findMethod(blockClass, GETMETANAME_METHOD);
                writeMethod = findMethod(blockClass, WRITE_METHOD,
                        OutputStream.class);

                block = constructor.newInstance(sfg.getJDKSignatureFile(), /*
                                                                             * explicit
                                                                             * argument
                                                                             * on
                                                                             * the
                                                                             * constructor
                                                                             */
                privateKey, certChain, externalSF, null, null, null, null,
                        zipFile);
            }

            public String getMetaName() throws IllegalAccessException,
                    InvocationTargetException {
                return (String) getMetaNameMethod.invoke(block);
            }

            public void write(OutputStream os) throws IllegalAccessException,
                    InvocationTargetException {
                writeMethod.invoke(block, os);
            }
        }
    }

    public static JarSigner newInstance(String keystoreLocation, String alias,
            String passwd) throws Exception {
        FileInputStream fileIn = new FileInputStream(keystoreLocation);
        KeyStore keyStore = KeyStore.getInstance("JKS");
        keyStore.load(fileIn, passwd.toCharArray());
        return newInstance(keyStore, alias, passwd);
    }

    public static JarSigner newInstance(KeyStore keyStore, String alias,
            String passwd) throws Exception {
        java.security.cert.Certificate[] chain = keyStore
                .getCertificateChain(alias);
        X509Certificate certChain[] = new X509Certificate[chain.length];

        CertificateFactory cf = CertificateFactory.getInstance("X.509");
        for (int count = 0; count < chain.length; count++) {
            ByteArrayInputStream certIn = new ByteArrayInputStream(chain[count]
                    .getEncoded());
            X509Certificate cert = (X509Certificate) cf
                    .generateCertificate(certIn);
            certChain[count] = cert;
        }

        Key key = keyStore.getKey(alias, passwd.toCharArray());
        KeyFactory keyFactory = KeyFactory.getInstance(key.getAlgorithm());
        KeySpec keySpec = keyFactory.getKeySpec(key, RSAPrivateKeySpec.class);
        PrivateKey privateKey = keyFactory.generatePrivate(keySpec);

        return new JarSigner(alias, privateKey, certChain);
    }
}
